from typing import Sequence

from faran.types import NumPyCosts, CostSumFunction

import numpy as np


class NumPyCostSumFunction[CostsT: NumPyCosts](CostSumFunction[CostsT]):
    """Sums multiple cost arrays element-wise into a single cost."""

    def __call__(self, costs: Sequence[CostsT], *, initial: CostsT) -> CostsT:
        return initial.similar(
            array=np.sum([it.array for it in costs], axis=0) + initial.array
        )



from typing import Protocol, NamedTuple, runtime_checkable, cast

from faran.types import (
    Array,
    NumPyGaussianBelief,
    NumPyNoiseCovariances,
    NumPyNoiseModelProvider,
)
from faran.filters.noise import IdentityNoiseModelProvider
from faran.filters.kf import numpy_kalman_filter

from jaxtyping import Float, Int

import numpy as np


@runtime_checkable
class StateTransitionFunction(Protocol):
    def __call__(self, state: Float[Array, "D_x K"]) -> Float[Array, "D_x K"]:
        """Applies the state transition function to the state."""
        ...


class NumPyUnscentedKalmanFilter(NamedTuple):
    """Unscented Kalman Filter for nonlinear systems with linear observations."""

    alpha: float
    beta: float
    noise_model: NumPyNoiseModelProvider

    @staticmethod
    def create(
        *,
        alpha: float = 1.0,
        beta: float = 2.0,
        noise_model: NumPyNoiseModelProvider | None = None,
    ) -> "NumPyUnscentedKalmanFilter":
        """Create an Unscented Kalman Filter.

        Args:
            alpha: Controls the spread of sigma points around the mean.
            beta: Incorporates prior knowledge (2 is optimal for Gaussian).
            noise_model: An optional noise model provider for adaptive noise estimation.
        """
        return NumPyUnscentedKalmanFilter(
            alpha=alpha,
            beta=beta,
            noise_model=IdentityNoiseModelProvider()
            if noise_model is None
            else noise_model,
        )

    def filter(
        self,
        observations: Float[Array, "T D_z K"],
        *,
        initial_state_covariance: Float[Array, "D_x D_x"],
        state_transition: StateTransitionFunction,
        process_noise_covariance: Float[Array, "D_x D_x"],
        observation_noise_covariance: Float[Array, "D_z D_z"],
        observation_matrix: Float[Array, "D_z D_x"],
    ) -> NumPyGaussianBelief:
        """Run the UKF over a sequence of observations.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Σ₀ matrix representing initial state uncertainty.
            state_transition: Nonlinear state transition function.
            process_noise_covariance: R matrix representing the covariance of process noise.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            observation_matrix: H matrix mapping state to observation space.
        """
        belief = self.initial_belief_from(
            observations, initial_state_covariance=initial_state_covariance
        )
        noise = NumPyNoiseCovariances(
            process_noise_covariance, observation_noise_covariance
        )
        adapt = self.noise_model(
            obstacle_count=observations.shape[2],
            observation_matrix=observation_matrix,
            noise=noise,
        )
        noise_state = adapt.state

        for observation in observations:
            belief = self.predict(
                belief=belief,
                state_transition=state_transition,
                process_noise_covariance=noise.process_noise_covariance,
            )

            noise, noise_state = adapt(
                noise=noise,
                prediction=belief,
                observation=observation,
                state=noise_state,
            )

            belief = self.update(
                observation=observation,
                prediction=belief,
                observation_matrix=observation_matrix,
                observation_noise_covariance=noise.observation_noise_covariance,
                initial_state_covariance=initial_state_covariance,
            )

        return cast(NumPyGaussianBelief, belief)

    def predict(
        self,
        *,
        belief: NumPyGaussianBelief,
        state_transition: StateTransitionFunction,
        process_noise_covariance: Float[Array, "D_x D_x"],
    ) -> NumPyGaussianBelief:
        """Performs the prediction step of the UKF using the unscented transform.

        Args:
            belief: The current belief about the state.
            state_transition: Nonlinear state transition function.
            process_noise_covariance: R matrix representing the covariance of process noise.
        """
        R = process_noise_covariance
        mu, sigma = belief

        n, obstacle_count = mu.shape
        lambda_ = self.scaling_parameter_for(n)
        sigma_point_count = 2 * n + 1

        mean_weights, covariance_weights = self._compute_weights(n, lambda_)

        def propagate(
            sigma_points: Float[Array, "K N_SigmaPoints D_x"],
        ) -> Float[Array, "K N_SigmaPoints D_x"]:
            all_points = sigma_points.reshape(-1, n).T
            propagated_flat = state_transition(all_points)  # type: ignore
            return propagated_flat.T.reshape(-1, sigma_point_count, n)  # type: ignore

        def reconstruct_belief(
            propagated: Float[Array, "K N_SigmaPoints D_x"],
        ) -> tuple[Float[Array, "K D_x"], Float[Array, "K D_x D_x"]]:
            means = np.sum(
                mean_weights[np.newaxis, :, np.newaxis] * propagated,
                axis=1,
            )

            deviations = propagated - means[:, np.newaxis, :]
            covariances = (
                np.einsum("s,ksi,ksj->kij", covariance_weights, deviations, deviations)
                + R
            )

            return means, covariances

        def scatter_to_full(
            *,
            means: Float[Array, "K_valid D_x"],
            covariances: Float[Array, "K_valid D_x D_x"],
            valid_indices: Int[Array, " K_valid"],
        ) -> NumPyGaussianBelief:
            predicted_mean = np.full_like(mu, np.nan)
            predicted_covariance = np.full_like(sigma, np.nan)
            predicted_mean[:, valid_indices] = means.T  # type: ignore
            predicted_covariance[:, :, valid_indices] = covariances.transpose(1, 2, 0)  # type: ignore
            return NumPyGaussianBelief(
                mean=predicted_mean, covariance=predicted_covariance
            )

        valid_indices = np.where(~np.any(np.isnan(sigma), axis=(0, 1)))[0]

        if len(valid_indices) == 0:
            return NumPyGaussianBelief(
                mean=np.full_like(mu, np.nan),
                covariance=np.full_like(sigma, np.nan),
            )

        sigma_points = self._generate_sigma_points_batch(
            means=mu[:, valid_indices],
            covariances=sigma[:, :, valid_indices],
            lambda_=lambda_,
        )
        propagated = propagate(sigma_points)
        means, covariances = reconstruct_belief(propagated)

        return scatter_to_full(
            means=means, covariances=covariances, valid_indices=valid_indices
        )

    def update(
        self,
        observation: Float[Array, "D_z K"],
        *,
        prediction: NumPyGaussianBelief,
        observation_matrix: Float[Array, "D_z D_x"],
        observation_noise_covariance: Float[Array, "D_z D_z"],
        initial_state_covariance: Float[Array, "D_x D_x"],
    ) -> NumPyGaussianBelief:
        """Performs the update step of the UKF using a linear observation model.

        Args:
            observation: The newly observed state.
            prediction: The predicted belief from the prediction step.
            observation_matrix: H matrix mapping state to observation space.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return numpy_kalman_filter.update(
            observation=observation,
            prediction=prediction,
            observation_matrix=observation_matrix,
            observation_noise_covariance=observation_noise_covariance,
            initial_state_covariance=initial_state_covariance,
        )

    def initial_belief_from(
        self,
        observations: Float[Array, "T D_z K"],
        *,
        initial_state_covariance: Float[Array, "D_x D_x"],
    ) -> NumPyGaussianBelief:
        """Initializes the belief state from the first observation using a pseudo-inverse.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return numpy_kalman_filter.initial_belief_from(
            observations, initial_state_covariance=initial_state_covariance
        )

    def scaling_parameter_for(self, state_dimension: int) -> float:
        """Returns the scaling parameter for the given state dimension."""
        return (self.alpha**2 - 1) * state_dimension

    def _compute_weights(
        self, state_dimension: int, lambda_: float
    ) -> tuple[Float[Array, " N_SigmaPoints"], Float[Array, " N_SigmaPoints"]]:
        n = state_dimension
        sigma_point_count = 2 * n + 1

        mean_weights = np.zeros(sigma_point_count)
        covariance_weights = np.zeros(sigma_point_count)

        mean_weights[0] = lambda_ / (n + lambda_)
        covariance_weights[0] = lambda_ / (n + lambda_) + (
            1 - self.alpha**2 + self.beta
        )

        remaining_weight = 1 / (2 * (n + lambda_))
        mean_weights[1:] = remaining_weight
        covariance_weights[1:] = remaining_weight

        return mean_weights, covariance_weights

    def _generate_sigma_points_batch(
        self,
        *,
        means: Float[Array, "D_x K"],
        covariances: Float[Array, "D_x D_x K"],
        lambda_: float,
    ) -> Float[Array, "K N_SigmaPoints D_x"]:
        n, obstacle_count = means.shape
        sigma_point_count = 2 * n + 1

        batched_covariance = covariances.transpose(2, 0, 1)
        scaled = (n + lambda_) * batched_covariance

        # NOTE: To ensure symmetric positive definite
        scaled = (scaled + scaled.transpose(0, 2, 1)) / 2
        eigenvalues, eigenvectors = np.linalg.eigh(scaled)
        eigenvalues = np.maximum(eigenvalues, 1e-10)
        scaled = eigenvectors @ (
            eigenvalues[..., np.newaxis] * eigenvectors.transpose(0, 2, 1)
        )

        sqrt_covariance = np.linalg.cholesky(scaled)
        means_t = means.T

        # NOTE: Use columns of L (transposed to row-major) for sigma point offsets
        sqrt_covariance_columns = sqrt_covariance.transpose(0, 2, 1)

        sigma_points = np.empty((obstacle_count, sigma_point_count, n))
        sigma_points[:, 0, :] = means_t
        sigma_points[:, 1 : n + 1, :] = (
            means_t[:, np.newaxis, :] + sqrt_covariance_columns
        )
        sigma_points[:, n + 1 :, :] = (
            means_t[:, np.newaxis, :] - sqrt_covariance_columns
        )

        return sigma_points



from typing import cast
from dataclasses import dataclass

from faran.types import JaxObstacleStateSampler
from faran.obstacles.accelerated import JaxSampledObstacle2dPoses, JaxObstacle2dPoses

from jaxtyping import PRNGKeyArray
from riskit import distribution

import jax
import jax.numpy as jnp


@dataclass
class JaxGaussianObstacle2dPoseSampler(
    JaxObstacleStateSampler[JaxObstacle2dPoses, JaxSampledObstacle2dPoses]
):
    """Samples obstacle poses from a Gaussian distribution parameterized by predicted covariances."""

    key: PRNGKeyArray

    @staticmethod
    def create(*, seed: int = 42) -> "JaxGaussianObstacle2dPoseSampler":
        return JaxGaussianObstacle2dPoseSampler(key=jax.random.key(seed))

    def __call__(
        self, states: JaxObstacle2dPoses, *, count: int
    ) -> JaxSampledObstacle2dPoses:
        if states.count == 0:
            return cast(JaxSampledObstacle2dPoses, states.single())

        if (covariance := states.covariance_array) is None:
            assert count == 1, (
                "It's pointless to take multiple samples, when covariance information is not available."
            )
            return cast(JaxSampledObstacle2dPoses, states.single())

        T, D_O, _, K = covariance.shape

        mean = jnp.stack([states.x_array, states.y_array, states.heading_array], axis=1)
        flat_covariance = covariance.transpose(0, 3, 1, 2).reshape(-1, D_O, D_O)
        flat_mean = mean.transpose(0, 2, 1).reshape(-1, D_O)

        gaussian = distribution.jax.gaussian(
            mean=flat_mean, covariance=flat_covariance, key=self.key
        )
        samples = gaussian.sample(count=count)
        self.key = gaussian.key

        samples = samples.reshape(T, K, D_O, count).transpose(0, 2, 1, 3)

        return states.sampled(
            x=samples[:, 0, :, :], y=samples[:, 1, :, :], heading=samples[:, 2, :, :]
        )



from typing import Self, Sequence, cast
from dataclasses import dataclass

from faran.types import (
    DataType,
    Array,
    NumPyState,
    NumPyStateSequence,
    NumPyStateBatch,
    NumPyControlInputSequence,
    NumPyControlInputBatch,
    AugmentedState,
    AugmentedStateSequence,
    AugmentedStateBatch,
    AugmentedControlInputSequence,
    AugmentedControlInputBatch,
    HasPhysical,
    HasVirtual,
    StateSequenceCreator,
)
from faran.states.augmented.base import (
    BaseAugmentedState,
    BaseAugmentedStateSequence,
    BaseAugmentedStateBatch,
    BaseAugmentedControlInputSequence,
    BaseAugmentedControlInputBatch,
)

from jaxtyping import Float

import numpy as np


@dataclass(frozen=True)
class NumPyAugmentedState[P: NumPyState, V: NumPyState](
    AugmentedState[P, V], HasPhysical[P], HasVirtual[V], NumPyState
):
    """NumPy augmented state pairing a physical and a virtual state."""

    inner: BaseAugmentedState[P, V]

    @staticmethod
    def of[P_: NumPyState, V_: NumPyState](
        *, physical: P_, virtual: V_
    ) -> "NumPyAugmentedState[P_, V_]":
        return NumPyAugmentedState(
            BaseAugmentedState.of(physical=physical, virtual=virtual)
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, " D_x"]:
        return np.asarray(self.inner)

    @property
    def physical(self) -> P:
        return self.inner.physical

    @property
    def virtual(self) -> V:
        return self.inner.virtual

    @property
    def dimension(self) -> int:
        return self.inner.dimension

    @property
    def array(self) -> Float[Array, " D_x"]:
        return np.concatenate(
            [self.inner.physical.array, self.inner.virtual.array], axis=0
        )


@dataclass(frozen=True)
class NumPyAugmentedStateSequence[P: NumPyStateSequence, V: NumPyStateSequence](
    AugmentedStateSequence[P, V],
    HasPhysical[P],
    HasVirtual[V],
    NumPyStateSequence,
):
    """NumPy augmented state sequence pairing physical and virtual sequences."""

    inner: BaseAugmentedStateSequence[P, V]

    @staticmethod
    def of[P_: NumPyStateSequence, V_: NumPyStateSequence](
        *, physical: P_, virtual: V_
    ) -> "NumPyAugmentedStateSequence[P_, V_]":
        return NumPyAugmentedStateSequence(
            BaseAugmentedStateSequence.of(physical=physical, virtual=virtual)
        )

    @staticmethod
    def of_states[
        PS: NumPyState,
        PSS: NumPyStateSequence,
        VS: NumPyState,
        VSS: NumPyStateSequence,
    ](
        *,
        physical: StateSequenceCreator[PS, PSS],
        virtual: StateSequenceCreator[VS, VSS],
    ) -> StateSequenceCreator[
        NumPyAugmentedState[PS, VS], "NumPyAugmentedStateSequence[PSS, VSS]"
    ]:
        """Returns a state sequence creator for augmented states."""

        def creator(
            states: Sequence[NumPyAugmentedState[PS, VS]],
        ) -> "NumPyAugmentedStateSequence[PSS, VSS]":
            return NumPyAugmentedStateSequence.of(
                physical=physical([s.physical for s in states]),
                virtual=virtual([s.virtual for s in states]),
            )

        return creator

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_x"]:
        return self.inner.__array__(dtype=dtype)

    def batched(self) -> "NumPyAugmentedStateBatch ":
        return NumPyAugmentedStateBatch(self.inner.batched())

    @property
    def physical(self) -> P:
        return self.inner.physical

    @property
    def virtual(self) -> V:
        return self.inner.virtual

    @property
    def horizon(self) -> int:
        return self.inner.horizon

    @property
    def dimension(self) -> int:
        return self.inner.dimension

    @property
    def array(self) -> Float[Array, "T D_x"]:
        return np.concatenate(
            [self.inner.physical.array, self.inner.virtual.array], axis=1
        )


@dataclass(frozen=True)
class NumPyAugmentedStateBatch[P: NumPyStateBatch, V: NumPyStateBatch](
    AugmentedStateBatch[P, V], HasPhysical[P], HasVirtual[V], NumPyStateBatch
):
    """NumPy augmented state batch pairing physical and virtual batches."""

    inner: BaseAugmentedStateBatch[P, V]

    @staticmethod
    def of[P_: NumPyStateBatch, V_: NumPyStateBatch](
        *, physical: P_, virtual: V_
    ) -> "NumPyAugmentedStateBatch[P_, V_]":
        return NumPyAugmentedStateBatch(
            BaseAugmentedStateBatch.of(physical=physical, virtual=virtual)
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_x M"]:
        return self.inner.__array__(dtype=dtype)

    @property
    def physical(self) -> P:
        return self.inner.physical

    @property
    def virtual(self) -> V:
        return self.inner.virtual

    @property
    def horizon(self) -> int:
        return self.inner.horizon

    @property
    def dimension(self) -> int:
        return self.inner.dimension

    @property
    def rollout_count(self) -> int:
        return self.inner.rollout_count

    @property
    def array(self) -> Float[Array, "T D_x M"]:
        return np.concatenate(
            [self.inner.physical.array, self.inner.virtual.array], axis=1
        )


@dataclass(frozen=True)
class NumPyAugmentedControlInputSequence[
    P: NumPyControlInputSequence,
    V: NumPyControlInputSequence,
](
    AugmentedControlInputSequence[P, V],
    HasPhysical[P],
    HasVirtual[V],
    NumPyControlInputSequence,
):
    """NumPy augmented control input sequence pairing physical and virtual inputs."""

    inner: BaseAugmentedControlInputSequence[P, V]

    @staticmethod
    def of[P_: NumPyControlInputSequence, V_: NumPyControlInputSequence](
        *, physical: P_, virtual: V_
    ) -> "NumPyAugmentedControlInputSequence[P_, V_]":
        return NumPyAugmentedControlInputSequence(
            BaseAugmentedControlInputSequence.of(physical=physical, virtual=virtual)
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_u"]:
        return self.inner.__array__(dtype=dtype)

    def similar(self, *, array: Float[Array, "T D_u"]) -> Self:
        return self.__class__(
            cast(
                BaseAugmentedControlInputSequence[P, V],
                BaseAugmentedControlInputSequence.of(
                    physical=self.inner.physical.similar(
                        array=array[:, : self.inner.physical.dimension]
                    ),
                    virtual=self.inner.virtual.similar(
                        array=array[:, -self.inner.virtual.dimension :]
                    ),
                ),
            )
        )

    @property
    def physical(self) -> P:
        return self.inner.physical

    @property
    def virtual(self) -> V:
        return self.inner.virtual

    @property
    def horizon(self) -> int:
        return self.inner.horizon

    @property
    def dimension(self) -> int:
        return self.inner.dimension

    @property
    def array(self) -> Float[Array, "T D_u"]:
        return np.concatenate(
            [self.inner.physical.array, self.inner.virtual.array], axis=1
        )


@dataclass(frozen=True)
class NumPyAugmentedControlInputBatch[
    P: NumPyControlInputBatch,
    V: NumPyControlInputBatch,
](
    AugmentedControlInputBatch[P, V],
    HasPhysical[P],
    HasVirtual[V],
    NumPyControlInputBatch,
):
    """NumPy augmented control input batch pairing physical and virtual batches."""

    inner: BaseAugmentedControlInputBatch[P, V]

    @staticmethod
    def of[P_: NumPyControlInputBatch, V_: NumPyControlInputBatch](
        *, physical: P_, virtual: V_
    ) -> "NumPyAugmentedControlInputBatch[P_, V_]":
        return NumPyAugmentedControlInputBatch(
            BaseAugmentedControlInputBatch.of(physical=physical, virtual=virtual)
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_u M"]:
        return self.inner.__array__(dtype=dtype)

    @property
    def physical(self) -> P:
        return self.inner.physical

    @property
    def virtual(self) -> V:
        return self.inner.virtual

    @property
    def horizon(self) -> int:
        return self.inner.horizon

    @property
    def dimension(self) -> int:
        return self.inner.dimension

    @property
    def rollout_count(self) -> int:
        return self.inner.rollout_count

    @property
    def array(self) -> Float[Array, "T D_u M"]:
        return np.concatenate(
            [self.inner.physical.array, self.inner.virtual.array], axis=1
        )



from typing import Protocol

from faran.types.predictors import ObstacleStatesHistory
from faran.types.models.integrator.common import (
    IntegratorState,
    IntegratorStateSequence,
    IntegratorStateBatch,
    IntegratorControlInputSequence,
    IntegratorControlInputBatch,
)

from jaxtyping import Array as JaxArray, Float


class JaxIntegratorState(IntegratorState, Protocol):
    @property
    def array(self) -> Float[JaxArray, " D_x"]:
        """Returns the underlying JAX array representing the integrator state."""
        ...


class JaxIntegratorStateSequence(IntegratorStateSequence, Protocol):
    @property
    def array(self) -> Float[JaxArray, "T D_x"]:
        """Returns the underlying JAX array representing the integrator state sequence."""
        ...


class JaxIntegratorStateBatch(IntegratorStateBatch, Protocol):
    @property
    def array(self) -> Float[JaxArray, "T D_x M"]:
        """Returns the underlying JAX array representing the integrator state batch."""
        ...


class JaxIntegratorControlInputSequence(IntegratorControlInputSequence, Protocol):
    @property
    def array(self) -> Float[JaxArray, "T D_u"]:
        """Returns the underlying JAX array representing the integrator control input sequence."""
        ...


class JaxIntegratorControlInputBatch(IntegratorControlInputBatch, Protocol):
    @property
    def array(self) -> Float[JaxArray, "T D_u M"]:
        """Returns the underlying JAX array representing the integrator control input batch."""
        ...


class JaxIntegratorObstacleStatesHistory(ObstacleStatesHistory, Protocol):
    @property
    def array(self) -> Float[JaxArray, "T D_o K"]:
        """Returns the obstacle history as a JAX array."""
        ...



from typing import Self, Sequence, Protocol
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    DataType,
    Array,
    jaxtyped,
    ControlInputSequence,
    JaxState,
    JaxStateSequence,
    JaxStateBatch,
    JaxControlInputSequence,
    JaxControlInputBatch,
    JaxCosts,
    JaxSampledObstacleStates,
    JaxObstacleStates,
    JaxObstacleStatesForTimeStep,
)
from faran.states.simple.basic import NumPySimpleObstacleStatesForTimeStep

from jaxtyping import Array as JaxArray, Float

import numpy as np
import jax.numpy as jnp


class JaxControlInputSequenceLike(ControlInputSequence, Protocol):
    """Protocol for JAX control input sequences exposing an array property."""

    @property
    def array(self) -> Float[JaxArray, "T D_u"]:
        """Returns the underlying JAX array representing the control input sequence."""
        ...


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleState(JaxState):
    """JAX simple flat state vector."""

    _array: Float[JaxArray, " D_x"]

    @staticmethod
    def create(
        array: Float[Array, " D_x"] | Float[JaxArray, " D_x"],
    ) -> "JaxSimpleState":
        """Creates a JAX simple state from the given array."""
        return JaxSimpleState(jnp.asarray(array))

    @staticmethod
    def zeroes(*, dimension: int) -> "JaxSimpleState":
        """Creates a zeroed simple state for the given dimension."""
        return JaxSimpleState(jnp.zeros((dimension,)))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, " D_x"]:
        return self._numpy_array

    @property
    def dimension(self) -> int:
        return self.array.shape[0]

    @property
    def array(self) -> Float[JaxArray, " D_x"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, " D_x"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleStateSequence(JaxStateSequence):
    """JAX state sequence as a 2D array (time × dimension)."""

    _array: Float[JaxArray, "T D_x"]

    @staticmethod
    def of_states(states: Sequence[JaxSimpleState]) -> "JaxSimpleStateSequence":
        """Creates a simple state sequence from a sequence of simple states."""
        assert len(states) > 0, "States sequence must not be empty."

        array = jnp.stack([state.array for state in states], axis=0)

        return JaxSimpleStateSequence(array)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_x"]:
        return self._numpy_array

    def batched(self) -> "JaxSimpleStateBatch":
        return JaxSimpleStateBatch.wrap(array=self.array[..., jnp.newaxis])

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[JaxArray, "T D_x"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T D_x"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleStateBatch(JaxStateBatch):
    """JAX state batch as a 3D array (time × dimension × rollouts)."""

    _array: Float[JaxArray, "T D_x M"]

    @staticmethod
    def wrap(
        array: Float[Array, "T D_x M"] | Float[JaxArray, "T D_x M"],
    ) -> "JaxSimpleStateBatch":
        """Creates a JAX simple state batch from the given array."""
        return JaxSimpleStateBatch(jnp.asarray(array))

    @staticmethod
    def of_states(states: Sequence[JaxState]) -> "JaxSimpleStateBatch":
        """Creates a simple state batch from a sequence of simple states."""
        assert len(states) > 0, "States sequence must not be empty."

        array = jnp.stack([state.array for state in states], axis=0)[:, :, jnp.newaxis]

        return JaxSimpleStateBatch(array)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_x M"]:
        return self._numpy_array

    def rollout(self, index: int) -> JaxSimpleStateSequence:
        return JaxSimpleStateSequence(self.array[..., index])

    def at(self, *, time_step: int, rollout: int) -> JaxSimpleState:
        return JaxSimpleState(self.array[time_step, :, rollout])

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> int:
        return self.array.shape[1]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> Float[JaxArray, "T D_x M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T D_x M"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleControlInputSequence(JaxControlInputSequence):
    """JAX control input sequence as a 2D array (time × dimension)."""

    _array: Float[JaxArray, "T D_u"]

    @staticmethod
    def create(
        array: Float[Array, "T D_u"] | Float[JaxArray, "T D_u"],
    ) -> "JaxSimpleControlInputSequence":
        """Creates a JAX simple control input sequence from the given array."""
        return JaxSimpleControlInputSequence(jnp.asarray(array))

    @staticmethod
    def constant(
        value: Float[JaxArray, " D_u"] | Float[Array, " D_u"], *, horizon: int
    ) -> "JaxSimpleControlInputSequence":
        """Creates a constant control input sequence for the given horizon and dimension."""
        return JaxSimpleControlInputSequence(
            jnp.tile(jnp.asarray(value)[jnp.newaxis, :], (horizon, 1))
        )

    @staticmethod
    def zeroes(*, horizon: int, dimension: int) -> "JaxSimpleControlInputSequence":
        """Creates a zeroed control input sequence for the given horizon."""
        return JaxSimpleControlInputSequence.constant(
            jnp.zeros((dimension,)), horizon=horizon
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_u"]:
        return self._numpy_array

    def similar(self, *, array: Float[JaxArray, "L D_u"]) -> Self:
        return self.__class__(array)

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[JaxArray, "T D_u"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T D_u"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleControlInputBatch(JaxControlInputBatch):
    """JAX control input batch as a 3D array (time × dimension × rollouts)."""

    _array: Float[JaxArray, "T D_u M"]

    @staticmethod
    def create(
        array: Float[Array, "T D_u M"] | Float[JaxArray, "T D_u M"],
    ) -> "JaxSimpleControlInputBatch":
        """Factory method to create a JAX simple control input batch from an array."""
        return JaxSimpleControlInputBatch(jnp.asarray(array))

    @staticmethod
    def zero(
        *, horizon: int, dimension: int, rollout_count: int = 1
    ) -> "JaxSimpleControlInputBatch":
        """Creates a zeroed control input batch for the given horizon and rollout count."""
        return JaxSimpleControlInputBatch(
            jnp.zeros((horizon, dimension, rollout_count))
        )

    @staticmethod
    def of(sequence: JaxControlInputSequenceLike) -> "JaxSimpleControlInputBatch":
        """Creates a JAX simple control input batch from a single control input sequence."""
        return JaxSimpleControlInputBatch(sequence.array[..., jnp.newaxis])

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_u M"]:
        return self._numpy_array

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> int:
        return self.array.shape[1]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> Float[JaxArray, "T D_u M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T D_u M"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleCosts(JaxCosts):
    """JAX cost values as a 2D array (time × rollouts)."""

    _array: Float[JaxArray, "T M"]

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T M"]:
        return self._numpy_array

    def similar(self, *, array: Float[JaxArray, "T M"]) -> Self:
        return self.__class__(array)

    def zero(self) -> Self:
        return self.__class__(jnp.zeros_like(self.array))

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[JaxArray, "T M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T M"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleSampledObstacleStates(JaxSampledObstacleStates):
    """JAX sampled obstacle states as a 4D array (time × dimension × obstacles × samples)."""

    _array: Float[JaxArray, "T D_o K N"]

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_o K N"]:
        return self._numpy_array

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> int:
        return self.array.shape[1]

    @property
    def count(self) -> int:
        return self.array.shape[2]

    @property
    def sample_count(self) -> int:
        return self.array.shape[3]

    @property
    def array(self) -> Float[JaxArray, "T D_o K N"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T D_o K N"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleObstacleStatesForTimeStep(
    JaxObstacleStatesForTimeStep[
        "JaxSimpleObstacleStates", NumPySimpleObstacleStatesForTimeStep
    ]
):
    """JAX obstacle states for a single time step as a 2D array (dimension × obstacles)."""

    _array: Float[JaxArray, "D_o K"]

    @staticmethod
    def create(
        array: Float[Array, "D_o K"] | Float[JaxArray, "D_o K"],
    ) -> "JaxSimpleObstacleStatesForTimeStep":
        """Creates obstacle states for a single time step from the given array."""
        return JaxSimpleObstacleStatesForTimeStep(jnp.asarray(array))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "D_o K"]:
        return self._numpy_array

    def replicate(self, *, horizon: int) -> "JaxSimpleObstacleStates":
        raise NotImplementedError(
            "Currently cannot be triggered by any component or test. Please "
            "implement once this functionality is actually needed."
        )

    def numpy(self) -> NumPySimpleObstacleStatesForTimeStep:
        return NumPySimpleObstacleStatesForTimeStep(self._numpy_array)

    @property
    def dimension(self) -> int:
        return self.array.shape[0]

    @property
    def count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[JaxArray, "D_o K"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "D_o K"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleObstacleStates(JaxObstacleStates[JaxSimpleSampledObstacleStates]):
    """JAX obstacle states as a 3D array (time × dimension × obstacles)."""

    _array: Float[JaxArray, "T D_o K"]
    _covariance: Float[JaxArray, "T D_o D_o K"] | None = None

    @staticmethod
    def create(
        *,
        states: Float[Array, "T D_o K"] | Float[JaxArray, "T D_o K"],
        covariance: Float[Array, "T D_o D_o K"]
        | Float[JaxArray, "T D_o D_o K"]
        | None = None,
    ) -> "JaxSimpleObstacleStates":

        return JaxSimpleObstacleStates(
            jnp.asarray(states),
            jnp.asarray(covariance) if covariance is not None else None,
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_o K"]:
        return self._numpy_array

    def single(self) -> JaxSimpleSampledObstacleStates:
        return JaxSimpleSampledObstacleStates(self.array[..., jnp.newaxis])

    def last(self) -> JaxSimpleObstacleStatesForTimeStep:
        return JaxSimpleObstacleStatesForTimeStep(self.array[-1])

    def covariance(self) -> Float[Array, "T D_o D_o K"] | None:
        return self._numpy_covariance

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> int:
        return self.array.shape[1]

    @property
    def count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> Float[JaxArray, "T D_o K"]:
        return self._array

    @property
    def covariance_array(self) -> Float[JaxArray, "T D_o D_o K"] | None:
        return self._covariance

    @cached_property
    def _numpy_array(self) -> Float[Array, "T D_o K"]:
        return np.asarray(self.array)

    @cached_property
    def _numpy_covariance(self) -> Float[Array, "T D_o D_o K"] | None:
        return np.asarray(self._covariance) if self._covariance is not None else None



from typing import Final, Any

from faran.types import (
    NumPyState,
    NumPyStateSequence,
    NumPyStateBatch,
    NumPyControlInputSequence,
    NumPyControlInputBatch,
    NumPyCosts,
    NumPyCostFunction,
    JaxState,
    JaxStateSequence,
    JaxStateBatch,
    JaxControlInputSequence,
    JaxControlInputBatch,
    JaxCosts,
    JaxCostFunction,
    State as State,
    StateSequence as StateSequence,
    StateBatch as StateBatch,
    ControlInputSequence as ControlInputSequence,
    ControlInputBatch as ControlInputBatch,
    Costs as Costs,
    CostFunction as CostFunction,
    BICYCLE_D_X,
    BICYCLE_D_U,
    BicycleD_x,
    BicycleD_u,
    BicycleState,
    BicycleStateSequence,
    BicycleStateBatch,
    BicyclePositions,
    BicycleControlInputSequence,
    BicycleControlInputBatch,
    UNICYCLE_D_X,
    UNICYCLE_D_U,
    UnicycleD_x,
    UnicycleD_u,
    UnicycleState,
    UnicycleStateSequence,
    UnicycleStateBatch,
    UnicyclePositions,
    UnicycleControlInputSequence,
    UnicycleControlInputBatch,
    NumPyPathParameters,
    NumPyReferencePoints,
    NumPyPositions,
    NumPySimplePositions,
    NumPyHeadings,
    NumPyLateralPositions,
    NumPyLongitudinalPositions,
    NumPyNormals,
    NumPyRisk,
    JaxPathParameters,
    JaxReferencePoints,
    JaxPositions,
    JaxSimplePositions,
    JaxHeadings,
    JaxLateralPositions,
    JaxLongitudinalPositions,
    JaxNormals,
    JaxRisk,
    PoseD_o as PoseD_o_,
    POSE_D_O as POSE_D_O_,
    NumPyPathParameterExtractor,
    NumPyPathVelocityExtractor,
    NumPyPositionExtractor,
    NumPyDistanceExtractor,
    NumPyRiskMetric,
    JaxPathParameterExtractor,
    JaxPathVelocityExtractor,
    JaxPositionExtractor,
    JaxDistanceExtractor,
    JaxRiskMetric,
    Error as Error,  # NOTE: Aliased to workaround ruff bug.
    Risk as Risk,
    RiskMetric as RiskMetric,
    ContouringCost as ContouringCost,
    NumPySampledObstacleStates,
    NumPySampledObstaclePositions,
    NumPySampledObstacleHeadings,
    NumPySampledObstaclePositionExtractor,
    NumPySampledObstacleHeadingExtractor,
    NumPyObstacleStates,
    NumPyObstacleStatesForTimeStep,
    NumPyObstacleStateProvider,
    NumPyObstaclePositionExtractor,
    JaxSampledObstacleStates,
    JaxSampledObstaclePositions,
    JaxSampledObstacleHeadings,
    JaxSampledObstaclePositionExtractor,
    JaxSampledObstacleHeadingExtractor,
    JaxObstacleStates,
    JaxObstacleStatesForTimeStep,
    JaxObstacleStateProvider,
    JaxObstaclePositionExtractor,
    AugmentedState,
    AugmentedStateSequence,
    AugmentedStateBatch,
    AugmentedControlInputSequence,
    AugmentedControlInputBatch,
    NumPyBoundaryDistance,
    NumPyBoundaryDistanceExtractor,
    JaxBoundaryDistance,
    JaxBoundaryDistanceExtractor,
)
from faran.models import (
    NumPyBicycleState,
    NumPyBicycleStateSequence,
    NumPyBicycleStateBatch,
    NumPyBicyclePositions,
    NumPyBicycleControlInputSequence,
    NumPyBicycleControlInputBatch,
    NumPyBicycleObstacleStates,
    NumPyBicycleObstacleInputs,
    NumPyBicycleObstacleStateSequences,
    NumPyIntegratorObstacleStateSequences,
    JaxBicycleState,
    JaxBicycleStateSequence,
    JaxBicycleStateBatch,
    JaxBicyclePositions,
    JaxBicycleControlInputSequence,
    JaxBicycleControlInputBatch,
    JaxBicycleObstacleStates,
    JaxBicycleObstacleInputs,
    JaxBicycleObstacleStateSequences,
    JaxIntegratorObstacleStateSequences,
    NumPyUnicycleState,
    NumPyUnicycleStateSequence,
    NumPyUnicycleStateBatch,
    NumPyUnicyclePositions,
    NumPyUnicycleControlInputSequence,
    NumPyUnicycleControlInputBatch,
    NumPyUnicycleObstacleStates,
    NumPyUnicycleObstacleInputs,
    NumPyUnicycleObstacleStateSequences,
    JaxUnicycleState,
    JaxUnicycleStateSequence,
    JaxUnicycleStateBatch,
    JaxUnicyclePositions,
    JaxUnicycleControlInputSequence,
    JaxUnicycleControlInputBatch,
    JaxUnicycleObstacleStates,
    JaxUnicycleObstacleInputs,
    JaxUnicycleObstacleStateSequences,
)
from faran.costs import (
    NumPyContouringCost,
    JaxContouringCost,
    NumPyDistance,
    JaxDistance,
)
from faran.obstacles import (
    NumPyObstacleIds,
    NumPySampledObstacle2dPoses,
    NumPyObstacle2dPoses,
    NumPyObstacle2dPosesForTimeStep,
    NumPyObstacle2dPositions,
    NumPyObstacle2dPositionsForTimeStep,
    NumPyObstacleHeadings,
    NumPyObstacleHeadingsForTimeStep,
    NumPyObstacleStatesRunningHistory,
    JaxObstacleIds,
    JaxSampledObstacle2dPoses,
    JaxObstacle2dPoses,
    JaxObstacle2dPosesForTimeStep,
    JaxObstacle2dPositions,
    JaxObstacle2dPositionsForTimeStep,
    JaxObstacleHeadings,
    JaxObstacleHeadingsForTimeStep,
    JaxObstacleStatesRunningHistory,
)
from faran.states import (
    NumPySimpleState,
    NumPySimpleStateSequence,
    NumPySimpleStateBatch,
    NumPySimpleControlInputSequence,
    NumPySimpleControlInputBatch,
    NumPySimpleCosts,
    NumPySimpleSampledObstacleStates,
    NumPySimpleObstacleStatesForTimeStep,
    NumPySimpleObstacleStates,
    JaxSimpleState,
    JaxSimpleStateSequence,
    JaxSimpleStateBatch,
    JaxSimpleControlInputSequence,
    JaxSimpleControlInputBatch,
    JaxSimpleCosts,
    JaxSimpleSampledObstacleStates,
    JaxSimpleObstacleStatesForTimeStep,
    JaxSimpleObstacleStates,
    NumPyAugmentedState,
    NumPyAugmentedStateSequence,
    NumPyAugmentedStateBatch,
    NumPyAugmentedControlInputSequence,
    NumPyAugmentedControlInputBatch,
    JaxAugmentedState,
    JaxAugmentedStateSequence,
    JaxAugmentedStateBatch,
    JaxAugmentedControlInputSequence,
    JaxAugmentedControlInputBatch,
)


class types:
    """Namespace of type aliases for states, controls, costs, and related domain types."""

    type State = State
    type StateBatch = StateBatch
    type ControlInputSequence = ControlInputSequence
    type ControlInputBatch = ControlInputBatch
    type Costs = Costs
    type CostFunction[InputBatchT, StateBatchT, CostsT] = CostFunction[
        InputBatchT, StateBatchT, CostsT
    ]
    type Error = Error
    type Risk = Risk

    type ContouringCost[InputBatchT, StateBatchT, ErrorT] = ContouringCost[
        InputBatchT, StateBatchT, ErrorT
    ]
    type RiskMetric[CostFunctionT, StateBatchT, ObstacleStatesT, SamplerT, RiskT] = (
        RiskMetric[CostFunctionT, StateBatchT, ObstacleStatesT, SamplerT, RiskT]
    )

    class obstacle:
        type PoseD_o = PoseD_o_

        POSE_D_O: Final = POSE_D_O_

    class bicycle:
        type D_x = BicycleD_x
        type D_u = BicycleD_u
        type State = BicycleState
        type StateSequence = BicycleStateSequence
        type StateBatch = BicycleStateBatch
        type Positions = BicyclePositions
        type ControlInputSequence = BicycleControlInputSequence
        type ControlInputBatch = BicycleControlInputBatch

        D_X: Final = BICYCLE_D_X
        D_U: Final = BICYCLE_D_U

    class unicycle:
        type D_x = UnicycleD_x
        type D_u = UnicycleD_u
        type State = UnicycleState
        type StateSequence = UnicycleStateSequence
        type StateBatch = UnicycleStateBatch
        type Positions = UnicyclePositions
        type ControlInputSequence = UnicycleControlInputSequence
        type ControlInputBatch = UnicycleControlInputBatch

        D_X: Final = UNICYCLE_D_X
        D_U: Final = UNICYCLE_D_U

    class augmented:
        type State[P, V] = AugmentedState[P, V]
        type StateSequence[P, V] = AugmentedStateSequence[P, V]
        type StateBatch[P, V] = AugmentedStateBatch[P, V]
        type ControlInputSequence[P, V] = AugmentedControlInputSequence[P, V]
        type ControlInputBatch[P, V] = AugmentedControlInputBatch[P, V]

        state: Final = AugmentedState
        state_sequence: Final = AugmentedStateSequence
        state_batch: Final = AugmentedStateBatch
        control_input_sequence: Final = AugmentedControlInputSequence
        control_input_batch: Final = AugmentedControlInputBatch

    class numpy:
        type State = NumPyState
        type StateSequence = NumPyStateSequence
        type StateBatch = NumPyStateBatch
        type ControlInputSequence = NumPyControlInputSequence
        type ControlInputBatch = NumPyControlInputBatch
        type Costs = NumPyCosts
        type PathParameters = NumPyPathParameters
        type ReferencePoints = NumPyReferencePoints
        type Positions = NumPyPositions
        type Headings = NumPyHeadings
        type LateralPositions = NumPyLateralPositions
        type LongitudinalPositions = NumPyLongitudinalPositions
        type ObstacleIds = NumPyObstacleIds
        type SampledObstacleStates = NumPySampledObstacleStates
        type SampledObstaclePositions = NumPySampledObstaclePositions
        type SampledObstacleHeadings = NumPySampledObstacleHeadings
        type SampledObstaclePositionExtractor[SampledStatesT] = (
            NumPySampledObstaclePositionExtractor[SampledStatesT]
        )
        type SampledObstacleHeadingExtractor[SampledStatesT] = (
            NumPySampledObstacleHeadingExtractor[SampledStatesT]
        )
        type ObstacleStates[SingleSampleT = Any, ObstacleStatesForTimeStepT = Any] = (
            NumPyObstacleStates[SingleSampleT, ObstacleStatesForTimeStepT]
        )
        type ObstacleStatesForTimeStep[ObstacleStatesT = Any] = (
            NumPyObstacleStatesForTimeStep[ObstacleStatesT]
        )
        type ObstaclePositionExtractor[
            ObstacleStatesForTimeStepT,
            ObstacleStatesT,
            PositionsForTimeStepT,
            PositionsT,
        ] = NumPyObstaclePositionExtractor[
            ObstacleStatesForTimeStepT,
            ObstacleStatesT,
            PositionsForTimeStepT,
            PositionsT,
        ]
        type SampledObstacle2dPoses = NumPySampledObstacle2dPoses
        type Obstacle2dPoses = NumPyObstacle2dPoses
        type Obstacle2dPosesForTimeStep = NumPyObstacle2dPosesForTimeStep
        type Obstacle2dPositions = NumPyObstacle2dPositions
        type Obstacle2dPositionsForTimeStep = NumPyObstacle2dPositionsForTimeStep
        type ObstacleHeadings = NumPyObstacleHeadings
        type ObstacleHeadingsForTimeStep = NumPyObstacleHeadingsForTimeStep
        type ObstacleStatesRunningHistory[
            StatesT,
            StatesForTimeStepT: NumPyObstacleStatesForTimeStep,
        ] = NumPyObstacleStatesRunningHistory[StatesT, StatesForTimeStepT]
        type Distance = NumPyDistance
        type BoundaryDistance = NumPyBoundaryDistance
        type Risk = NumPyRisk

        type CostFunction = NumPyCostFunction
        type PathParameterExtractor[StateBatchT] = NumPyPathParameterExtractor[
            StateBatchT
        ]
        type PathVelocityExtractor[InputBatchT] = NumPyPathVelocityExtractor[
            InputBatchT
        ]
        type PositionExtractor[StateBatchT] = NumPyPositionExtractor[StateBatchT]
        type DistanceExtractor[StateBatchT, SampledObstacleStatesT, DistanceT] = (
            NumPyDistanceExtractor[StateBatchT, SampledObstacleStatesT, DistanceT]
        )
        type BoundaryDistanceExtractor[StateBatchT, DistanceT] = (
            NumPyBoundaryDistanceExtractor[StateBatchT, DistanceT]
        )
        type RiskMetric[StateBatchT, ObstacleStatesT, SampledObstacleStatesT] = (
            NumPyRiskMetric[StateBatchT, ObstacleStatesT, SampledObstacleStatesT]
        )
        type ContouringCost[StateBatchT] = NumPyContouringCost[StateBatchT]
        type ObstacleStateProvider[ObstacleStatesT] = NumPyObstacleStateProvider[
            ObstacleStatesT
        ]

        path_parameters: Final = NumPyPathParameters
        reference_points: Final = NumPyReferencePoints.create
        positions: Final = NumPySimplePositions.create
        headings: Final = NumPyHeadings.create
        lateral_positions: Final = NumPyLateralPositions.create
        longitudinal_positions: Final = NumPyLongitudinalPositions.create
        normals: Final = NumPyNormals.create
        distance: Final = NumPyDistance
        boundary_distance: Final = NumPyBoundaryDistance
        obstacle_ids: Final = NumPyObstacleIds
        obstacle_2d_poses: Final = NumPyObstacle2dPoses
        obstacle_2d_poses_for_time_step: Final = NumPyObstacle2dPosesForTimeStep
        obstacle_states_running_history: Final = NumPyObstacleStatesRunningHistory

        class simple:
            type State = NumPySimpleState
            type StateSequence = NumPySimpleStateSequence
            type StateBatch = NumPySimpleStateBatch
            type ControlInputSequence = NumPySimpleControlInputSequence
            type ControlInputBatch = NumPySimpleControlInputBatch
            type Costs = NumPySimpleCosts
            type SampledObstacleStates = NumPySimpleSampledObstacleStates
            type ObstacleStatesForTimeStep = NumPySimpleObstacleStatesForTimeStep
            type ObstacleStates = NumPySimpleObstacleStates

            state: Final = NumPySimpleState
            state_sequence: Final = NumPySimpleStateSequence
            state_batch: Final = NumPySimpleStateBatch
            control_input_sequence: Final = NumPySimpleControlInputSequence
            control_input_batch: Final = NumPySimpleControlInputBatch
            costs: Final = NumPySimpleCosts
            obstacle_states: Final = NumPySimpleObstacleStates
            obstacle_states_for_time_step: Final = NumPySimpleObstacleStatesForTimeStep

        class integrator:
            type ObstacleStateSequences = NumPyIntegratorObstacleStateSequences

        class bicycle:
            type State = NumPyBicycleState
            type StateSequence = NumPyBicycleStateSequence
            type StateBatch = NumPyBicycleStateBatch
            type Positions = NumPyBicyclePositions
            type ControlInputSequence = NumPyBicycleControlInputSequence
            type ControlInputBatch = NumPyBicycleControlInputBatch
            type ObstacleStates = NumPyBicycleObstacleStates
            type ObstacleInputs = NumPyBicycleObstacleInputs
            type ObstacleStateSequences = NumPyBicycleObstacleStateSequences

            state: Final = NumPyBicycleState
            state_sequence: Final = NumPyBicycleStateSequence
            state_batch: Final = NumPyBicycleStateBatch
            positions: Final = NumPyBicyclePositions
            control_input_sequence: Final = NumPyBicycleControlInputSequence
            control_input_batch: Final = NumPyBicycleControlInputBatch
            obstacle_states: Final = NumPyBicycleObstacleStates
            obstacle_inputs: Final = NumPyBicycleObstacleInputs
            obstacle_state_sequences: Final = NumPyBicycleObstacleStateSequences

        class unicycle:
            type State = NumPyUnicycleState
            type StateSequence = NumPyUnicycleStateSequence
            type StateBatch = NumPyUnicycleStateBatch
            type Positions = NumPyUnicyclePositions
            type ControlInputSequence = NumPyUnicycleControlInputSequence
            type ControlInputBatch = NumPyUnicycleControlInputBatch
            type ObstacleStates = NumPyUnicycleObstacleStates
            type ObstacleInputs = NumPyUnicycleObstacleInputs
            type ObstacleStateSequences = NumPyUnicycleObstacleStateSequences

            state: Final = NumPyUnicycleState
            state_sequence: Final = NumPyUnicycleStateSequence
            state_batch: Final = NumPyUnicycleStateBatch
            positions: Final = NumPyUnicyclePositions
            control_input_sequence: Final = NumPyUnicycleControlInputSequence
            control_input_batch: Final = NumPyUnicycleControlInputBatch
            obstacle_states: Final = NumPyUnicycleObstacleStates
            obstacle_inputs: Final = NumPyUnicycleObstacleInputs
            obstacle_state_sequences: Final = NumPyUnicycleObstacleStateSequences

        class augmented:
            type State[P: NumPyState, V: NumPyState] = NumPyAugmentedState[P, V]
            type StateSequence[P: NumPyStateSequence, V: NumPyStateSequence] = (
                NumPyAugmentedStateSequence[P, V]
            )
            type StateBatch[P: NumPyStateBatch, V: NumPyStateBatch] = (
                NumPyAugmentedStateBatch[P, V]
            )
            type ControlInputSequence[
                P: NumPyControlInputSequence,
                V: NumPyControlInputSequence,
            ] = NumPyAugmentedControlInputSequence[P, V]
            type ControlInputBatch[
                P: NumPyControlInputBatch,
                V: NumPyControlInputBatch,
            ] = NumPyAugmentedControlInputBatch[P, V]

            state: Final = NumPyAugmentedState
            state_sequence: Final = NumPyAugmentedStateSequence
            state_batch: Final = NumPyAugmentedStateBatch
            control_input_sequence: Final = NumPyAugmentedControlInputSequence
            control_input_batch: Final = NumPyAugmentedControlInputBatch

    class jax:
        type State = JaxState
        type StateSequence = JaxStateSequence
        type StateBatch = JaxStateBatch
        type ControlInputSequence = JaxControlInputSequence
        type ControlInputBatch = JaxControlInputBatch
        type Costs = JaxCosts
        type PathParameters = JaxPathParameters
        type ReferencePoints = JaxReferencePoints
        type Positions = JaxPositions
        type Headings = JaxHeadings
        type LateralPositions = JaxLateralPositions
        type LongitudinalPositions = JaxLongitudinalPositions
        type ObstacleIds = JaxObstacleIds
        type SampledObstacleStates = JaxSampledObstacleStates
        type SampledObstaclePositions = JaxSampledObstaclePositions
        type SampledObstacleHeadings = JaxSampledObstacleHeadings
        type SampledObstaclePositionExtractor[SampledStatesT] = (
            JaxSampledObstaclePositionExtractor[SampledStatesT]
        )
        type SampledObstacleHeadingExtractor[SampledStatesT] = (
            JaxSampledObstacleHeadingExtractor[SampledStatesT]
        )
        type ObstacleStates[SingleSampleT = Any, ObstacleStatesForTimeStepT = Any] = (
            JaxObstacleStates[SingleSampleT, ObstacleStatesForTimeStepT]
        )
        type ObstacleStatesForTimeStep[ObstacleStatesT = Any, NumPyT = Any] = (
            JaxObstacleStatesForTimeStep[ObstacleStatesT, NumPyT]
        )
        type ObstaclePositionExtractor[
            ObstacleStatesForTimeStepT,
            ObstacleStatesT,
            PositionsForTimeStepT,
            PositionsT,
        ] = JaxObstaclePositionExtractor[
            ObstacleStatesForTimeStepT,
            ObstacleStatesT,
            PositionsForTimeStepT,
            PositionsT,
        ]
        type SampledObstacle2dPoses = JaxSampledObstacle2dPoses
        type Obstacle2dPoses = JaxObstacle2dPoses
        type Obstacle2dPosesForTimeStep = JaxObstacle2dPosesForTimeStep
        type Obstacle2dPositions = JaxObstacle2dPositions
        type Obstacle2dPositionsForTimeStep = JaxObstacle2dPositionsForTimeStep
        type ObstacleHeadings = JaxObstacleHeadings
        type ObstacleHeadingsForTimeStep = JaxObstacleHeadingsForTimeStep
        type ObstacleStatesRunningHistory[
            StatesT: JaxObstacleStates,
            StatesForTimeStepT: JaxObstacleStatesForTimeStep,
        ] = JaxObstacleStatesRunningHistory[StatesT, StatesForTimeStepT]
        type Distance = JaxDistance
        type BoundaryDistance = JaxBoundaryDistance
        type Risk = JaxRisk

        type CostFunction = JaxCostFunction
        type PathParameterExtractor[StateBatchT] = JaxPathParameterExtractor[
            StateBatchT
        ]
        type PathVelocityExtractor[InputBatchT] = JaxPathVelocityExtractor[InputBatchT]
        type PositionExtractor[StateBatchT] = JaxPositionExtractor[StateBatchT]
        type DistanceExtractor[StateBatchT, SampledObstacleStatesT, DistanceT] = (
            JaxDistanceExtractor[StateBatchT, SampledObstacleStatesT, DistanceT]
        )
        type BoundaryDistanceExtractor[StateBatchT, DistanceT] = (
            JaxBoundaryDistanceExtractor[StateBatchT, DistanceT]
        )
        type RiskMetric[StateBatchT, ObstacleStatesT, SampledObstacleStatesT] = (
            JaxRiskMetric[StateBatchT, ObstacleStatesT, SampledObstacleStatesT]
        )
        type ContouringCost[StateBatchT] = JaxContouringCost[StateBatchT]
        type ObstacleStateProvider[ObstacleStatesT] = JaxObstacleStateProvider[
            ObstacleStatesT
        ]

        path_parameters: Final = JaxPathParameters.create
        reference_points: Final = JaxReferencePoints.create
        positions: Final = JaxSimplePositions.create
        headings: Final = JaxHeadings.create
        lateral_positions: Final = JaxLateralPositions.create
        longitudinal_positions: Final = JaxLongitudinalPositions.create
        normals: Final = JaxNormals.create
        distance: Final = JaxDistance
        boundary_distance: Final = JaxBoundaryDistance
        obstacle_ids: Final = JaxObstacleIds
        obstacle_2d_poses: Final = JaxObstacle2dPoses
        obstacle_2d_poses_for_time_step: Final = JaxObstacle2dPosesForTimeStep
        obstacle_states_running_history: Final = JaxObstacleStatesRunningHistory

        class simple:
            type State = JaxSimpleState
            type StateSequence = JaxSimpleStateSequence
            type StateBatch = JaxSimpleStateBatch
            type ControlInputSequence = JaxSimpleControlInputSequence
            type ControlInputBatch = JaxSimpleControlInputBatch
            type Costs = JaxSimpleCosts
            type SampledObstacleStates = JaxSimpleSampledObstacleStates
            type ObstacleStatesForTimeStep = JaxSimpleObstacleStatesForTimeStep
            type ObstacleStates = JaxSimpleObstacleStates

            state: Final = JaxSimpleState
            state_sequence: Final = JaxSimpleStateSequence
            state_batch: Final = JaxSimpleStateBatch
            control_input_sequence: Final = JaxSimpleControlInputSequence
            control_input_batch: Final = JaxSimpleControlInputBatch
            costs: Final = JaxSimpleCosts
            obstacle_states: Final = JaxSimpleObstacleStates
            obstacle_states_for_time_step: Final = JaxSimpleObstacleStatesForTimeStep

        class integrator:
            type ObstacleStateSequences = JaxIntegratorObstacleStateSequences

        class bicycle:
            type State = JaxBicycleState
            type StateSequence = JaxBicycleStateSequence
            type StateBatch = JaxBicycleStateBatch
            type Positions = JaxBicyclePositions
            type ControlInputSequence = JaxBicycleControlInputSequence
            type ControlInputBatch = JaxBicycleControlInputBatch
            type ObstacleStates = JaxBicycleObstacleStates
            type ObstacleInputs = JaxBicycleObstacleInputs
            type ObstacleStateSequences = JaxBicycleObstacleStateSequences

            state: Final = JaxBicycleState
            state_sequence: Final = JaxBicycleStateSequence
            state_batch: Final = JaxBicycleStateBatch
            positions: Final = JaxBicyclePositions
            control_input_sequence: Final = JaxBicycleControlInputSequence
            control_input_batch: Final = JaxBicycleControlInputBatch
            obstacle_states: Final = JaxBicycleObstacleStates
            obstacle_inputs: Final = JaxBicycleObstacleInputs
            obstacle_state_sequences: Final = JaxBicycleObstacleStateSequences

        class unicycle:
            type State = JaxUnicycleState
            type StateSequence = JaxUnicycleStateSequence
            type StateBatch = JaxUnicycleStateBatch
            type Positions = JaxUnicyclePositions
            type ControlInputSequence = JaxUnicycleControlInputSequence
            type ControlInputBatch = JaxUnicycleControlInputBatch
            type ObstacleStates = JaxUnicycleObstacleStates
            type ObstacleInputs = JaxUnicycleObstacleInputs
            type ObstacleStateSequences = JaxUnicycleObstacleStateSequences

            state: Final = JaxUnicycleState
            state_sequence: Final = JaxUnicycleStateSequence
            state_batch: Final = JaxUnicycleStateBatch
            positions: Final = JaxUnicyclePositions
            control_input_sequence: Final = JaxUnicycleControlInputSequence
            control_input_batch: Final = JaxUnicycleControlInputBatch
            obstacle_states: Final = JaxUnicycleObstacleStates
            obstacle_inputs: Final = JaxUnicycleObstacleInputs
            obstacle_state_sequences: Final = JaxUnicycleObstacleStateSequences

        class augmented:
            type State[P: JaxState, V: JaxState] = JaxAugmentedState[P, V]
            type StateSequence[P: JaxStateSequence, V: JaxStateSequence] = (
                JaxAugmentedStateSequence[P, V]
            )
            type StateBatch[P: JaxStateBatch, V: JaxStateBatch] = (
                JaxAugmentedStateBatch[P, V]
            )
            type ControlInputSequence[
                P: JaxControlInputSequence,
                V: JaxControlInputSequence,
            ] = JaxAugmentedControlInputSequence[P, V]
            type ControlInputBatch[P: JaxControlInputBatch, V: JaxControlInputBatch] = (
                JaxAugmentedControlInputBatch[P, V]
            )

            state: Final = JaxAugmentedState
            state_sequence: Final = JaxAugmentedStateSequence
            state_batch: Final = JaxAugmentedStateBatch
            control_input_sequence: Final = JaxAugmentedControlInputSequence
            control_input_batch: Final = JaxAugmentedControlInputBatch


class classes:
    """Namespace of concrete protocol classes for states, controls, costs, and related types."""

    State: Final = State
    StateSequence: Final = StateSequence
    StateBatch: Final = StateBatch
    ControlInputSequence: Final = ControlInputSequence
    ControlInputBatch: Final = ControlInputBatch
    Costs: Final = Costs
    CostFunction: Final = CostFunction
    Error: Final = Error
    Risk: Final = Risk
    RiskMetric: Final = RiskMetric
    ContouringCost: Final = ContouringCost

    class bicycle:
        State: Final = BicycleState
        StateSequence: Final = BicycleStateSequence
        StateBatch: Final = BicycleStateBatch
        Positions: Final = BicyclePositions
        ControlInputSequence: Final = BicycleControlInputSequence
        ControlInputBatch: Final = BicycleControlInputBatch

    class unicycle:
        State: Final = UnicycleState
        StateSequence: Final = UnicycleStateSequence
        StateBatch: Final = UnicycleStateBatch
        Positions: Final = UnicyclePositions
        ControlInputSequence: Final = UnicycleControlInputSequence
        ControlInputBatch: Final = UnicycleControlInputBatch

    class augmented:
        State: Final = AugmentedState
        StateSequence: Final = AugmentedStateSequence
        StateBatch: Final = AugmentedStateBatch
        ControlInputSequence: Final = AugmentedControlInputSequence
        ControlInputBatch: Final = AugmentedControlInputBatch

    class numpy:
        State: Final = NumPyState
        StateSequence: Final = NumPyStateSequence
        StateBatch: Final = NumPyStateBatch
        ControlInputSequence: Final = NumPyControlInputSequence
        ControlInputBatch: Final = NumPyControlInputBatch
        Costs: Final = NumPyCosts
        PathParameters: Final = NumPyPathParameters
        ReferencePoints: Final = NumPyReferencePoints
        Positions: Final = NumPyPositions
        Headings: Final = NumPyHeadings
        LateralPositions: Final = NumPyLateralPositions
        LongitudinalPositions: Final = NumPyLongitudinalPositions
        ObstacleIds: Final = NumPyObstacleIds
        ObstacleStates: Final = NumPyObstacleStates
        ObstacleStatesForTimeStep: Final = NumPyObstacleStatesForTimeStep
        SampledObstacleStates: Final = NumPySampledObstacleStates
        SampledObstaclePositions: Final = NumPySampledObstaclePositions
        SampledObstacleHeadings: Final = NumPySampledObstacleHeadings
        SampledObstaclePositionExtractor: Final = NumPySampledObstaclePositionExtractor
        SampledObstacleHeadingExtractor: Final = NumPySampledObstacleHeadingExtractor
        SampledObstacle2dPoses: Final = NumPySampledObstacle2dPoses
        Obstacle2dPoses: Final = NumPyObstacle2dPoses
        Obstacle2dPosesForTimeStep: Final = NumPyObstacle2dPosesForTimeStep
        Obstacle2dPositions: Final = NumPyObstacle2dPositions
        Obstacle2dPositionsForTimeStep: Final = NumPyObstacle2dPositionsForTimeStep
        ObstacleHeadings: Final = NumPyObstacleHeadings
        ObstacleHeadingsForTimeStep: Final = NumPyObstacleHeadingsForTimeStep
        ObstacleStatesRunningHistory: Final = NumPyObstacleStatesRunningHistory
        ObstaclePositionExtractor: Final = NumPyObstaclePositionExtractor
        Distance: Final = NumPyDistance
        BoundaryDistance: Final = NumPyBoundaryDistance
        Risk: Final = NumPyRisk
        CostFunction: Final = NumPyCostFunction
        PathParameterExtractor: Final = NumPyPathParameterExtractor
        PathVelocityExtractor: Final = NumPyPathVelocityExtractor
        PositionExtractor: Final = NumPyPositionExtractor
        DistanceExtractor: Final = NumPyDistanceExtractor
        BoundaryDistanceExtractor: Final = NumPyBoundaryDistanceExtractor
        RiskMetric: Final = NumPyRiskMetric
        ContouringCost: Final = NumPyContouringCost
        ObstacleStateProvider: Final = NumPyObstacleStateProvider

        class simple:
            State: Final = NumPySimpleState
            StateSequence: Final = NumPySimpleStateSequence
            StateBatch: Final = NumPySimpleStateBatch
            ControlInputSequence: Final = NumPySimpleControlInputSequence
            ControlInputBatch: Final = NumPySimpleControlInputBatch
            Costs: Final = NumPySimpleCosts
            SampledObstacleStates: Final = NumPySimpleSampledObstacleStates
            ObstacleStatesForTimeStep: Final = NumPySimpleObstacleStatesForTimeStep
            ObstacleStates: Final = NumPySimpleObstacleStates

        class integrator:
            ObstacleStateSequences: Final = NumPyIntegratorObstacleStateSequences

        class bicycle:
            State: Final = NumPyBicycleState
            StateSequence: Final = NumPyBicycleStateSequence
            StateBatch: Final = NumPyBicycleStateBatch
            Positions: Final = NumPyBicyclePositions
            ControlInputSequence: Final = NumPyBicycleControlInputSequence
            ControlInputBatch: Final = NumPyBicycleControlInputBatch
            ObstacleStateSequences: Final = NumPyBicycleObstacleStateSequences

        class unicycle:
            State: Final = NumPyUnicycleState
            StateSequence: Final = NumPyUnicycleStateSequence
            StateBatch: Final = NumPyUnicycleStateBatch
            Positions: Final = NumPyUnicyclePositions
            ControlInputSequence: Final = NumPyUnicycleControlInputSequence
            ControlInputBatch: Final = NumPyUnicycleControlInputBatch
            ObstacleStateSequences: Final = NumPyUnicycleObstacleStateSequences

        class augmented:
            State: Final = NumPyAugmentedState
            StateSequence: Final = NumPyAugmentedStateSequence
            StateBatch: Final = NumPyAugmentedStateBatch
            ControlInputSequence: Final = NumPyAugmentedControlInputSequence
            ControlInputBatch: Final = NumPyAugmentedControlInputBatch

    class jax:
        State: Final = JaxState
        StateSequence: Final = JaxStateSequence
        StateBatch: Final = JaxStateBatch
        ControlInputSequence: Final = JaxControlInputSequence
        ControlInputBatch: Final = JaxControlInputBatch
        Costs: Final = JaxCosts
        PathParameters: Final = JaxPathParameters
        ReferencePoints: Final = JaxReferencePoints
        Positions: Final = JaxPositions
        Headings: Final = JaxHeadings
        LateralPositions: Final = JaxLateralPositions
        LongitudinalPositions: Final = JaxLongitudinalPositions
        ObstacleIds: Final = JaxObstacleIds
        ObstacleStates: Final = JaxObstacleStates
        ObstacleStatesForTimeStep: Final = JaxObstacleStatesForTimeStep
        SampledObstacleStates: Final = JaxSampledObstacleStates
        SampledObstaclePositions: Final = JaxSampledObstaclePositions
        SampledObstacleHeadings: Final = JaxSampledObstacleHeadings
        SampledObstaclePositionExtractor: Final = JaxSampledObstaclePositionExtractor
        SampledObstacleHeadingExtractor: Final = JaxSampledObstacleHeadingExtractor
        SampledObstacle2dPoses: Final = JaxSampledObstacle2dPoses
        Obstacle2dPoses: Final = JaxObstacle2dPoses
        Obstacle2dPosesForTimeStep: Final = JaxObstacle2dPosesForTimeStep
        Obstacle2dPositions: Final = JaxObstacle2dPositions
        Obstacle2dPositionsForTimeStep: Final = JaxObstacle2dPositionsForTimeStep
        ObstacleHeadings: Final = JaxObstacleHeadings
        ObstacleHeadingsForTimeStep: Final = JaxObstacleHeadingsForTimeStep
        ObstacleStatesRunningHistory: Final = JaxObstacleStatesRunningHistory
        ObstaclePositionExtractor: Final = JaxObstaclePositionExtractor
        Distance: Final = JaxDistance
        BoundaryDistance: Final = JaxBoundaryDistance
        Risk: Final = JaxRisk
        CostFunction: Final = JaxCostFunction
        PathParameterExtractor: Final = JaxPathParameterExtractor
        PathVelocityExtractor: Final = JaxPathVelocityExtractor
        PositionExtractor: Final = JaxPositionExtractor
        DistanceExtractor: Final = JaxDistanceExtractor
        BoundaryDistanceExtractor: Final = JaxBoundaryDistanceExtractor
        RiskMetric: Final = JaxRiskMetric
        ContouringCost: Final = JaxContouringCost
        ObstacleStateProvider: Final = JaxObstacleStateProvider

        class simple:
            State: Final = JaxSimpleState
            StateSequence: Final = JaxSimpleStateSequence
            StateBatch: Final = JaxSimpleStateBatch
            ControlInputSequence: Final = JaxSimpleControlInputSequence
            ControlInputBatch: Final = JaxSimpleControlInputBatch
            Costs: Final = JaxSimpleCosts
            SampledObstacleStates: Final = JaxSimpleSampledObstacleStates
            ObstacleStatesForTimeStep: Final = JaxSimpleObstacleStatesForTimeStep
            ObstacleStates: Final = JaxSimpleObstacleStates

        class integrator:
            ObstacleStateSequences: Final = JaxIntegratorObstacleStateSequences

        class bicycle:
            State: Final = JaxBicycleState
            StateSequence: Final = JaxBicycleStateSequence
            StateBatch: Final = JaxBicycleStateBatch
            Positions: Final = JaxBicyclePositions
            ControlInputSequence: Final = JaxBicycleControlInputSequence
            ControlInputBatch: Final = JaxBicycleControlInputBatch
            ObstacleStateSequences: Final = JaxBicycleObstacleStateSequences

        class unicycle:
            State: Final = JaxUnicycleState
            StateSequence: Final = JaxUnicycleStateSequence
            StateBatch: Final = JaxUnicycleStateBatch
            Positions: Final = JaxUnicyclePositions
            ControlInputSequence: Final = JaxUnicycleControlInputSequence
            ControlInputBatch: Final = JaxUnicycleControlInputBatch
            ObstacleStateSequences: Final = JaxUnicycleObstacleStateSequences

        class augmented:
            State: Final = JaxAugmentedState
            StateSequence: Final = JaxAugmentedStateSequence
            StateBatch: Final = JaxAugmentedStateBatch
            ControlInputSequence: Final = JaxAugmentedControlInputSequence
            ControlInputBatch: Final = JaxAugmentedControlInputBatch



from typing import Protocol, overload
from dataclasses import dataclass
from functools import cached_property

from faran.types.array import Array, jaxtyped
from faran.types.trajectories.common import (
    D_R,
    PathParameters,
    ReferencePoints,
    Positions,
    LateralPositions,
    LongitudinalPositions,
    Normals,
)

from jaxtyping import Array as JaxArray, Float

import jax.numpy as jnp
import numpy as np


@dataclass(frozen=True)
class JaxPathParameters(PathParameters):
    array: Float[JaxArray, "T M"]

    @overload
    @staticmethod
    def create(array: Float[Array, "T M"]) -> "JaxPathParameters":
        """Creates a JAX path parameters instance from a NumPy array."""
        ...

    @overload
    @staticmethod
    def create(array: Float[JaxArray, "T M"]) -> "JaxPathParameters":
        """Creates a JAX path parameters instance from a JAX array."""
        ...

    @staticmethod
    def create(
        array: Float[Array, "T M"] | Float[JaxArray, "T M"],
    ) -> "JaxPathParameters":
        return JaxPathParameters(array=jnp.asarray(array))

    def __array__(self) -> Float[Array, "T M"]:
        return np.asarray(self.array)

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[1]


class JaxPositions(Positions, Protocol):
    @property
    def x_array(self) -> Float[JaxArray, "T M"]:
        """Returns the x coordinates as a JAX array."""
        ...

    @property
    def y_array(self) -> Float[JaxArray, "T M"]:
        """Returns the y coordinates as a JAX array."""
        ...

    @property
    def array(self) -> Float[JaxArray, "T 2 M"]:
        """Returns the positions as a JAX array."""
        ...


@jaxtyped
@dataclass(frozen=True)
class JaxSimplePositions(JaxPositions):
    _x_array: Float[JaxArray, "T M"]
    _y_array: Float[JaxArray, "T M"]

    @staticmethod
    def create(
        *,
        x: Float[JaxArray, "T M"] | Float[Array, "T M"],
        y: Float[JaxArray, "T M"] | Float[Array, "T M"],
    ) -> "JaxSimplePositions":
        """Creates a JAX positions instance from x and y coordinate arrays."""
        return JaxSimplePositions(_x_array=jnp.asarray(x), _y_array=jnp.asarray(y))

    def __array__(self) -> Float[Array, "T 2 M"]:
        return self._numpy_array

    def x(self) -> Float[Array, "T M"]:
        return np.asarray(self._x_array)

    def y(self) -> Float[Array, "T M"]:
        return np.asarray(self._y_array)

    @property
    def x_array(self) -> Float[JaxArray, "T M"]:
        return self._x_array

    @property
    def y_array(self) -> Float[JaxArray, "T M"]:
        return self._y_array

    @property
    def horizon(self) -> int:
        return self._x_array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._x_array.shape[1]

    @property
    def array(self) -> Float[JaxArray, "T 2 M"]:
        return self._array

    @cached_property
    def _array(self) -> Float[JaxArray, "T 2 M"]:
        return jnp.stack([self._x_array, self._y_array], axis=1)

    @cached_property
    def _numpy_array(self) -> Float[Array, "T 2 M"]:
        return np.asarray(self._array)


@jaxtyped
@dataclass(frozen=True)
class JaxHeadings:
    _heading: Float[JaxArray, "T M"]

    @staticmethod
    def create(
        *, heading: Float[Array, "T M"] | Float[JaxArray, "T M"]
    ) -> "JaxHeadings":
        """Creates a JAX headings instance from an array of headings."""
        return JaxHeadings(jnp.asarray(heading))

    def __array__(self) -> Float[Array, "T M"]:
        return self._numpy_heading

    def heading(self) -> Float[Array, "T M"]:
        return self._numpy_heading

    @property
    def horizon(self) -> int:
        return self._heading.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._heading.shape[1]

    @property
    def heading_array(self) -> Float[JaxArray, "T M"]:
        return self._heading

    @cached_property
    def _numpy_heading(self) -> Float[Array, "T M"]:
        return np.asarray(self._heading)


@dataclass(frozen=True)
class JaxReferencePoints(ReferencePoints):
    array: Float[JaxArray, f"T {D_R} M"]

    @overload
    @staticmethod
    def create(
        *,
        x: Float[Array, "T M"],
        y: Float[Array, "T M"],
        heading: Float[Array, "T M"],
    ) -> "JaxReferencePoints":
        """Creates a JAX reference points instance from NumPy arrays."""
        ...

    @overload
    @staticmethod
    def create(
        *,
        x: Float[JaxArray, "T M"],
        y: Float[JaxArray, "T M"],
        heading: Float[JaxArray, "T M"],
    ) -> "JaxReferencePoints":
        """Creates a JAX reference points instance from JAX arrays."""
        ...

    @staticmethod
    def create(
        *,
        x: Float[Array, "T M"] | Float[JaxArray, "T M"],
        y: Float[Array, "T M"] | Float[JaxArray, "T M"],
        heading: Float[Array, "T M"] | Float[JaxArray, "T M"],
    ) -> "JaxReferencePoints":
        return JaxReferencePoints(array=jnp.stack([x, y, heading], axis=1))

    def __array__(self) -> Float[Array, f"T {D_R} M"]:
        return np.asarray(self.array)

    def x(self) -> Float[Array, "T M"]:
        return np.asarray(self.array[:, 0])

    def y(self) -> Float[Array, "T M"]:
        return np.asarray(self.array[:, 1])

    def heading(self) -> Float[Array, "T M"]:
        return np.asarray(self.array[:, 2])

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def x_array(self) -> Float[JaxArray, "T M"]:
        return self.array[:, 0]

    @property
    def y_array(self) -> Float[JaxArray, "T M"]:
        return self.array[:, 1]

    @property
    def heading_array(self) -> Float[JaxArray, "T M"]:
        return self.array[:, 2]

    @property
    def positions_array(self) -> Float[JaxArray, "T 2 M"]:
        return self.array[:, :2]


@jaxtyped
@dataclass(frozen=True)
class JaxLateralPositions(LateralPositions):
    _array: Float[JaxArray, "T M"]

    @staticmethod
    def create(
        array: Float[JaxArray, "T M"] | Float[Array, "T M"],
    ) -> "JaxLateralPositions":
        """Creates a JAX lateral positions instance from an array."""
        return JaxLateralPositions(jnp.asarray(array))

    def __array__(self) -> Float[Array, "T M"]:
        return self._numpy_array

    @property
    def horizon(self) -> int:
        return self._array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._array.shape[1]

    @property
    def array(self) -> Float[JaxArray, "T M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T M"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxLongitudinalPositions(LongitudinalPositions):
    _array: Float[JaxArray, "T M"]

    @staticmethod
    def create(
        array: Float[JaxArray, "T M"] | Float[Array, "T M"],
    ) -> "JaxLongitudinalPositions":
        """Creates a JAX longitudinal positions instance from an array."""
        return JaxLongitudinalPositions(jnp.asarray(array))

    def __array__(self) -> Float[Array, "T M"]:
        return self._numpy_array

    @property
    def horizon(self) -> int:
        return self._array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._array.shape[1]

    @property
    def array(self) -> Float[JaxArray, "T M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T M"]:
        return np.asarray(self.array)


@jaxtyped
@dataclass(frozen=True)
class JaxNormals(Normals):
    _array: Float[JaxArray, "T 2 M"]

    @staticmethod
    def create(
        *,
        x: Float[JaxArray, "T M"] | Float[Array, "T M"],
        y: Float[JaxArray, "T M"] | Float[Array, "T M"],
    ) -> "JaxNormals":
        """Creates a JAX normals instance from x and y coordinate arrays."""
        return JaxNormals(_array=jnp.stack([x, y], axis=1))

    def __array__(self) -> Float[Array, "T 2 M"]:
        return self._numpy_array

    def x(self) -> Float[Array, "T M"]:
        return np.asarray(self._array[:, 0])

    def y(self) -> Float[Array, "T M"]:
        return np.asarray(self._array[:, 1])

    @property
    def horizon(self) -> int:
        return self._array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._array.shape[2]

    @property
    def array(self) -> Float[JaxArray, "T 2 M"]:
        return self._array

    @cached_property
    def _numpy_array(self) -> Float[Array, "T 2 M"]:
        return np.asarray(self.array)



from .common import (
    ObstaclePositionsForTimeStep as ObstaclePositionsForTimeStep,
    ObstaclePositions as ObstaclePositions,
    ObstaclePositionExtractor as ObstaclePositionExtractor,
    ObstacleOrientationsForTimeStep as ObstacleOrientationsForTimeStep,
    ObstacleOrientations as ObstacleOrientations,
    ObstacleOrientationExtractor as ObstacleOrientationExtractor,
)
from .basic import (
    NumPyObstaclePositionsForTimeStep as NumPyObstaclePositionsForTimeStep,
    NumPyObstaclePositions as NumPyObstaclePositions,
    NumPyObstaclePositionExtractor as NumPyObstaclePositionExtractor,
    NumPyObstacleOrientationsForTimeStep as NumPyObstacleOrientationsForTimeStep,
    NumPyObstacleOrientations as NumPyObstacleOrientations,
    NumPyObstacleOrientationExtractor as NumPyObstacleOrientationExtractor,
)
from .accelerated import (
    JaxObstaclePositionsForTimeStep as JaxObstaclePositionsForTimeStep,
    JaxObstaclePositions as JaxObstaclePositions,
    JaxObstaclePositionExtractor as JaxObstaclePositionExtractor,
    JaxObstacleOrientationsForTimeStep as JaxObstacleOrientationsForTimeStep,
    JaxObstacleOrientations as JaxObstacleOrientations,
    JaxObstacleOrientationExtractor as JaxObstacleOrientationExtractor,
)



from typing import Protocol

from faran.types.array import Array
from faran.types.obstacles.history.common import (
    ObstaclePositionsForTimeStep,
    ObstaclePositions,
    ObstacleOrientationsForTimeStep,
    ObstacleOrientations,
    ObstaclePositionExtractor,
    ObstacleOrientationExtractor,
)

from jaxtyping import Float


class NumPyObstaclePositionsForTimeStep(ObstaclePositionsForTimeStep, Protocol):
    @property
    def array(self) -> Float[Array, "D_p K"]:
        """Returns the obstacle positions for a single time step as a NumPy array."""
        ...


class NumPyObstaclePositions(ObstaclePositions, Protocol):
    @property
    def array(self) -> Float[Array, "T D_p K"]:
        """Returns the obstacle positions as a NumPy array."""
        ...


class NumPyObstacleOrientationsForTimeStep(ObstacleOrientationsForTimeStep, Protocol):
    @property
    def array(self) -> Float[Array, "D_o K"]:
        """Returns the obstacle orientations for a single time step as a NumPy array."""
        ...


class NumPyObstacleOrientations(ObstacleOrientations, Protocol):
    @property
    def array(self) -> Float[Array, "T D_o K"]:
        """Returns the obstacle orientations as a NumPy array."""
        ...


class NumPyObstaclePositionExtractor[
    ObstacleStatesForTimeStepT,
    ObstacleStatesT,
    PositionsForTimeStepT,
    PositionsT,
](
    ObstaclePositionExtractor[
        ObstacleStatesForTimeStepT, ObstacleStatesT, PositionsForTimeStepT, PositionsT
    ],
    Protocol,
): ...


class NumPyObstacleOrientationExtractor[
    ObstacleStatesForTimeStepT,
    ObstacleStatesT,
    OrientationsForTimeStepT,
    OrientationsT,
](
    ObstacleOrientationExtractor[
        ObstacleStatesForTimeStepT,
        ObstacleStatesT,
        OrientationsForTimeStepT,
        OrientationsT,
    ],
    Protocol,
): ...



from typing import Protocol, Sequence, Mapping, Literal, Type, Any, overload
from dataclasses import dataclass

type NotRequired = Literal[False]
type Required = Literal[True]
type IsRequired = NotRequired | Required


@dataclass(frozen=True)
class SimulationDataAccessor[T, RequiredT: IsRequired]:
    key: str
    required: RequiredT
    _type: Type[T]

    @staticmethod
    def create[T_](
        default_type: Type[T_], *, key: str
    ) -> "SimulationDataAccessor[T_, NotRequired]":
        return SimulationDataAccessor(key=key, required=False, _type=default_type)

    def require(self) -> "SimulationDataAccessor[T, Required]":
        return SimulationDataAccessor(key=self.key, required=True, _type=self._type)

    def assume[T_](self, new_type: Type[T_]) -> "SimulationDataAccessor[T_, RequiredT]":
        return SimulationDataAccessor(
            key=self.key, required=self.required, _type=new_type
        )


@dataclass(kw_only=True, frozen=True)
class SimulationData:
    data: Mapping[str, Any]

    @staticmethod
    def create(data: Mapping[str, Any]) -> "SimulationData":
        return SimulationData(data=data)

    @overload
    def __call__[T](self, accessor: SimulationDataAccessor[T, Required]) -> T:
        """Retrieves the data for the specified accessor, ensuring it is present."""
        ...

    @overload
    def __call__[T](self, accessor: SimulationDataAccessor[T, NotRequired]) -> T | None:
        """Retrieves the data for the specified accessor, returning None if not present."""
        ...

    def __call__(self, accessor: SimulationDataAccessor) -> Any | None:
        assert accessor.key in self.data or not accessor.required, (
            f"Required simulation data '{accessor.key}' is missing."
        )

        return self.data.get(accessor.key)


class StateSequenceCreator[StateT, StateSequenceT](Protocol):
    def __call__(self, states: Sequence[StateT]) -> StateSequenceT:
        """Creates a state sequence from the provided sequence of states."""
        ...


class ObstacleStateSequencesCreator[
    ObstacleStatesForTimeStepT,
    ObstacleStateSequencesT,
](Protocol):
    def __call__(
        self, obstacle_states: Sequence[ObstacleStatesForTimeStepT]
    ) -> ObstacleStateSequencesT:
        """Creates obstacle state sequences from the provided sequence of obstacle states."""
        ...


class ObstacleStateObserver[ObstacleStatesForTimeStepT](Protocol):
    def observe(self, states: ObstacleStatesForTimeStepT) -> None:
        """Observes the specified obstacle states for a given time step."""
        ...



from typing import Protocol

from faran.types.array import Array

from jaxtyping import Float


class NumPyControlInputBatchCreator[
    InputBatchT,
](Protocol):
    def __call__(self, *, array: Float[Array, "T D_u M"]) -> InputBatchT:
        """Creates a control input batch from the given array."""
        ...



from typing import NamedTuple
from dataclasses import dataclass

from faran.types import (
    Array,
    NumPyGaussianBelief,
    NumPyNoiseCovariances,
    NumPyNoiseModelProvider,
    NumPyNoiseCovarianceDescription,
)
from faran.filters.noise import IdentityNoiseModelProvider

from jaxtyping import Float

import numpy as np


class ObstaclePartitioning(NamedTuple):
    should_update: Float[Array, " K"]
    should_initialize: Float[Array, " K"]


@dataclass(frozen=True)
class NumPyKalmanFilter:
    """Kalman Filter for linear systems."""

    noise_model: NumPyNoiseModelProvider

    @staticmethod
    def create(
        *, noise_model: NumPyNoiseModelProvider | None = None
    ) -> "NumPyKalmanFilter":
        """Creates a Kalman filter for linear systems.

        Args:
            noise_model: An optional noise model provider for adaptive noise estimation.
        """
        return NumPyKalmanFilter(
            noise_model=IdentityNoiseModelProvider()
            if noise_model is None
            else noise_model
        )

    def filter(
        self,
        observations: Float[Array, "T D_z K"],
        *,
        initial_state_covariance: Float[Array, "D_x D_x"],
        state_transition_matrix: Float[Array, "D_x D_x"],
        process_noise_covariance: Float[Array, "D_x D_x"],
        observation_noise_covariance: Float[Array, "D_z D_z"],
        observation_matrix: Float[Array, "D_z D_x"],
    ) -> NumPyGaussianBelief:
        """Performs one step of the Kalman filter given a new observation.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
            state_transition_matrix: A matrix for state prediction.
            process_noise_covariance: R matrix representing the covariance of process noise.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            observation_matrix: H matrix mapping state to observation space.
        """
        belief = self.initial_belief_from(
            observations, initial_state_covariance=initial_state_covariance
        )
        noise = NumPyNoiseCovariances(
            process_noise_covariance, observation_noise_covariance
        )
        adapt = self.noise_model(
            obstacle_count=observations.shape[2],
            observation_matrix=observation_matrix,
            noise=noise,
        )
        noise_state = adapt.state

        for observation in observations:
            belief = self.predict(
                belief=belief,
                state_transition_matrix=state_transition_matrix,
                process_noise_covariance=noise.process_noise_covariance,
            )

            noise, noise_state = adapt(
                noise=noise,
                prediction=belief,
                observation=observation,
                state=noise_state,
            )

            belief = self.update(
                observation=observation,
                prediction=belief,
                observation_matrix=observation_matrix,
                observation_noise_covariance=noise.observation_noise_covariance,
                initial_state_covariance=initial_state_covariance,
            )

        return belief

    def predict(
        self,
        *,
        belief: NumPyGaussianBelief,
        state_transition_matrix: Float[Array, "D_x D_x"],
        process_noise_covariance: Float[Array, "D_x D_x"],
    ) -> NumPyGaussianBelief:
        """Performs the prediction step of the Kalman filter from the existing belief.

        Args:
            belief: The current belief about the state.
            state_transition_matrix: A matrix for state prediction.
            process_noise_covariance: R matrix representing the covariance of process noise.
        """
        return numpy_kalman_filter.predict(
            belief=belief,
            state_transition_matrix=state_transition_matrix,
            process_noise_covariance=process_noise_covariance,
        )

    def update(
        self,
        observation: Float[Array, "D_z K"],
        *,
        prediction: NumPyGaussianBelief,
        observation_matrix: Float[Array, "D_z D_x"],
        observation_noise_covariance: Float[Array, "D_z D_z"],
        initial_state_covariance: Float[Array, "D_x D_x"],
    ) -> NumPyGaussianBelief:
        """Performs the update step of the Kalman filter using a new observation.

        Args:
            observation: The newly observed state.
            prediction: The predicted belief from the prediction step.
            observation_matrix: H matrix mapping state to observation space.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return numpy_kalman_filter.update(
            observation,
            prediction=prediction,
            observation_matrix=observation_matrix,
            observation_noise_covariance=observation_noise_covariance,
            initial_state_covariance=initial_state_covariance,
        )

    def initial_belief_from(
        self,
        observations: Float[Array, "T D_z K"],
        *,
        initial_state_covariance: Float[Array, "D_x D_x"],
    ) -> NumPyGaussianBelief:
        """Initializes the belief state from the first observation using a pseudo-inverse.

        This method provides a simple way to initialize the state mean based on the first
        observation, while using the provided initial state covariance for uncertainty.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return numpy_kalman_filter.initial_belief_from(
            observations, initial_state_covariance=initial_state_covariance
        )


class numpy_kalman_filter:
    @staticmethod
    def predict(
        *,
        belief: NumPyGaussianBelief,
        state_transition_matrix: Float[Array, "D_x D_x"],
        process_noise_covariance: Float[Array, "D_x D_x"],
    ) -> NumPyGaussianBelief:
        A = state_transition_matrix
        R = process_noise_covariance
        mu, sigma = belief

        sigma_t = sigma.transpose(2, 0, 1)
        predicted_covariance = (A @ sigma_t @ A.T).transpose(1, 2, 0) + R[
            ..., np.newaxis
        ]

        return NumPyGaussianBelief(mean=A @ mu, covariance=predicted_covariance)

    @staticmethod
    def update(
        observation: Float[Array, "D_z K"],
        *,
        prediction: NumPyGaussianBelief,
        observation_matrix: Float[Array, "D_z D_x"],
        observation_noise_covariance: Float[Array, "D_z D_z"],
        initial_state_covariance: Float[Array, "D_x D_x"],
    ) -> NumPyGaussianBelief:
        H = observation_matrix
        Q = observation_noise_covariance

        def partition(
            *, observation: Float[Array, "D_z K"], mean: Float[Array, "D_x K"]
        ) -> ObstaclePartitioning:
            observation_valid = ~np.any(np.isnan(observation), axis=0)
            prediction_valid = ~np.any(np.isnan(mean), axis=0)

            return ObstaclePartitioning(
                should_update=observation_valid & prediction_valid,
                should_initialize=observation_valid & ~prediction_valid,
            )

        def substitute_missing_values(
            *,
            observation: Float[Array, "D_z K"],
            mean: Float[Array, "D_x K"],
            covariance: Float[Array, "D_x D_x K"],
        ) -> tuple[
            Float[Array, "D_z K"], Float[Array, "D_x K"], Float[Array, "D_x D_x K"]
        ]:
            # NOTE: We temporarily replace NaNs with zeros to prevent errors.
            covariance_is_nan = np.any(np.isnan(covariance), axis=(0, 1))

            return (
                np.where(np.isnan(observation), 0.0, observation),
                np.where(np.isnan(mean), 0.0, mean),
                np.where(
                    covariance_is_nan[np.newaxis, np.newaxis, :],
                    initial_state_covariance[..., np.newaxis],
                    covariance,
                ),
            )

        def update(
            *,
            observation: Float[Array, "D_z K"],
            mean: Float[Array, "D_x K"],
            covariance: Float[Array, "D_x D_x K"],
        ) -> NumPyGaussianBelief:
            D_x = prediction.mean.shape[0]

            sigma_t = covariance.transpose(2, 0, 1)
            S_t = H @ sigma_t @ H.T + Q
            K_t = np.linalg.solve(S_t, H @ sigma_t).transpose(0, 2, 1)

            innovation = observation - H @ mean
            updated_mean = mean + (K_t @ innovation.T[:, :, np.newaxis]).squeeze(-1).T

            IKH = np.eye(D_x) - K_t @ H
            joseph_term = IKH @ sigma_t @ IKH.transpose(0, 2, 1)
            kalman_noise_term = K_t @ Q @ K_t.transpose(0, 2, 1)
            updated_covariance = (joseph_term + kalman_noise_term).transpose(1, 2, 0)

            return NumPyGaussianBelief(mean=updated_mean, covariance=updated_covariance)

        def initialize_from(observation: Float[Array, "D_z K"]) -> NumPyGaussianBelief:
            K = observation.shape[1]

            # NOTE: The mean state is the observation when available, 0 otherwise.
            return NumPyGaussianBelief(
                mean=np.linalg.pinv(H) @ observation,
                covariance=np.repeat(
                    initial_state_covariance[..., np.newaxis], K, axis=2
                ),
            )

        def blend(
            partitioning: ObstaclePartitioning,
            *,
            initial: NumPyGaussianBelief,
            prediction: NumPyGaussianBelief,
            update: NumPyGaussianBelief,
        ) -> NumPyGaussianBelief:
            return NumPyGaussianBelief(
                mean=np.where(
                    partitioning.should_update,
                    update.mean,
                    np.where(
                        partitioning.should_initialize, initial.mean, prediction.mean
                    ),
                ),
                covariance=np.where(
                    partitioning.should_update,
                    update.covariance,
                    np.where(
                        partitioning.should_initialize,
                        initial.covariance,
                        prediction.covariance,
                    ),
                ),
            )

        partitioning = partition(observation=observation, mean=prediction.mean)

        if np.all(partitioning.should_update):
            return update(
                observation=observation,
                mean=prediction.mean,
                covariance=prediction.covariance,
            )

        observation, mean, covariance = substitute_missing_values(
            observation=observation,
            mean=prediction.mean,
            covariance=prediction.covariance,
        )

        updated = update(observation=observation, mean=mean, covariance=covariance)

        if np.any(partitioning.should_initialize):
            initial = initialize_from(observation)
        else:
            initial = prediction

        return blend(
            partitioning, initial=initial, prediction=prediction, update=updated
        )

    @staticmethod
    def initial_belief_from(
        observations: Float[Array, "T D_z K"],
        *,
        initial_state_covariance: Float[Array, "D_x D_x"],
    ) -> NumPyGaussianBelief:
        D_x = initial_state_covariance.shape[0]
        K = observations.shape[2]

        mean = np.full((D_x, K), np.nan)
        covariance = np.full((D_x, D_x, K), np.nan)

        return NumPyGaussianBelief(mean=mean, covariance=covariance)

    @staticmethod
    def standardize_noise_covariance(
        covariance: NumPyNoiseCovarianceDescription, *, dimension: int | None = None
    ) -> Float[Array, "D_c D_c"]:
        if isinstance(covariance, (int, float)):
            assert dimension is not None, (
                "Dimension must be provided for scalar covariance."
            )
            return covariance * np.eye(dimension)

        match covariance.ndim:
            case 2:
                return covariance
            case 1:
                return np.diag(covariance)
            case _:
                assert False, f"Invalid covariance shape: {covariance.shape}"



from typing import Self
from dataclasses import dataclass

from faran.types import Array, jaxtyped, JaxObstacleSimulator
from faran.obstacles.accelerated import JaxObstacle2dPosesForTimeStep

from jaxtyping import Array as JaxArray, Float

import jax.numpy as jnp


@jaxtyped
@dataclass(frozen=True)
class JaxStaticObstacleSimulator(JaxObstacleSimulator[JaxObstacle2dPosesForTimeStep]):
    """Simulates stationary obstacles by replicating fixed positions over the horizon."""

    positions: Float[JaxArray, "K 2"]
    headings: Float[JaxArray, " K"]

    @staticmethod
    def empty() -> "JaxStaticObstacleSimulator":
        positions = jnp.empty((0, 2))

        return JaxStaticObstacleSimulator.create(positions=positions)

    @staticmethod
    def create(
        *,
        positions: Float[Array, "K 2"] | Float[JaxArray, "K 2"],
        headings: Float[Array, " K"] | Float[JaxArray, " K"] | None = None,
    ) -> "JaxStaticObstacleSimulator":
        count = positions.shape[0]
        headings = headings if headings is not None else jnp.zeros(shape=(count,))

        return JaxStaticObstacleSimulator(
            positions=jnp.asarray(positions), headings=jnp.asarray(headings)
        )

    def with_time_step_size(self, time_step_size: float) -> Self:
        # Time step does not matter.
        return self

    def step(self) -> JaxObstacle2dPosesForTimeStep:
        return JaxObstacle2dPosesForTimeStep.create(
            x=self.positions[:, 0], y=self.positions[:, 1], heading=self.headings
        )

    @property
    def obstacle_count(self) -> int:
        return self.positions.shape[0]
